package MineSweeper;

import java.awt.*;
import java.awt.event.*;
import java.util.Date;
import java.util.Random;

public class MineField extends Panel implements MouseListener {

  private Dimension field_dim = null;
  private int mine_dim = 0;
  private MineButton field[][];

  private float mine_ratio = 0.0f;
  private int mine_count = 0;
  private int total_mines = 0;
  private int squares_left = 0;

  private boolean lighted = false;
  private boolean user_lock = false;

  private MineFieldListener listener = null;
  private Font font = null;
  private MineButton mb;

  private Image flag = null;
  private Image mine = null;
  private Image mineX = null;

  private static final Point null_point = new Point();

  // Constructor
  public MineField(Dimension field_dim, int mine_dim, float mine_ratio, Font font, 
		   Image flag, Image mine, Image mineX, MineFieldListener listener) {
    super(new GridLayout(field_dim.height, field_dim.width));

    // Initialize the internal state
    this.field_dim = field_dim;
    this.mine_dim = mine_dim;
    this.mine_ratio = mine_ratio;
    this.font = font;
    this.flag = flag;
    this.mine = mine;
    this.mineX = mineX;
    this.listener = listener;

    // Create the MineField
    CreateField();
  }

  public MineField(Dimension field_dim, int mine_dim, float mine_ratio, Font font, 
		   Image flag, Image mine, Image mineX) {
    super(new GridLayout(field_dim.height, field_dim.width));

    // Initialize the internal state
    this.field_dim = field_dim;
    this.mine_dim = mine_dim;
    this.mine_ratio = mine_ratio;
    this.font = font;
    this.flag = flag;
    this.mine = mine;
    this.mineX = mineX;

    // Create the MineField
    CreateField();
  }

  // This method sets the mine field's listener
  public void SetListener(MineFieldListener listener) {
    this.listener = listener;
  }


  // This method performs the actual creation and assembly of the MineField
  private void CreateField() {
    Random random = new Random((new Date()).getTime());
    float random_num, tmp_num;
    SetSquaresLeft(field_dim.height * field_dim.width);
    total_mines = (int)((float)squares_left * mine_ratio);
    field = new MineButton[field_dim.width][field_dim.height];

    for (int y = 0; y < field_dim.height; ++y) {
      for (int x = 0; x < field_dim.width; ++x) {
        random_num = random.nextFloat();
        tmp_num = ((float)(total_mines-GetMineCount()))/((float)squares_left);
        field[x][y] = new MineButton(new Point(x,y), random_num < tmp_num, mine_dim, 
				     font);
        field[x][y].addMouseListener(this);
        this.add(field[x][y]);
        if (random_num < tmp_num) ++mine_count;
        --squares_left;
      }
    }
    squares_left = field_dim.height * field_dim.width;
  }

  // This method performs the actual creation and assembly of the MineField
  public void ResetField() {
    Random random = new Random((new Date()).getTime());
    float random_num, tmp_num;
    squares_left = field_dim.height * field_dim.width;
    mine_count = 0;
    user_lock = false;

    for (int y = 0; y < field_dim.height; ++y) {
      for (int x = 0; x < field_dim.width; ++x) {
        random_num = random.nextFloat();
        tmp_num = ((float)(total_mines-mine_count))/((float)squares_left);
        field[x][y].Reset(random_num < tmp_num);
        if (random_num < tmp_num) ++mine_count;
        --squares_left;
      }
    }
    SetMineCount(mine_count);
    SetSquaresLeft(field_dim.height * field_dim.width);
  }

  public void ClickFree() {
    for (int y = 0; y < field_dim.height; ++y) {
      for (int x = 0; x < field_dim.width; ++x) {
        if (!field[x][y].IsHiddenMine() && (MineCount(field[x][y]) == 0) && 
	    (field[x][y].IsFree())) {
          CheckAt(field[x][y]);
	  return;
        }
      }
    }
  }

  // These methods simply update the mine_count and notify the listener
  private void SetMineCount(int mine_count) {
    this.mine_count = mine_count;
    if (listener != null)
      listener.NewMineCount(mine_count);
  }

  private void IncMineCount() {
    ++mine_count;
    if (listener != null)
      listener.NewMineCount(mine_count);
  }

  private void DecMineCount() {
    --mine_count;
    if (listener != null)
      listener.NewMineCount(mine_count);
  }

  public int GetMineCount() {
    return mine_count;
  }

  public void UserDead() {
    if (listener != null)
      listener.UserDead();
  }

  public void UserLock() {
    user_lock = true;
    if (listener != null)
      listener.UserLocked();
  }


  // These methods update the squares_left variable and notify the
  // listener of the changes
  private void SetSquaresLeft(int squares_left) {
    this.squares_left = squares_left;
    if (listener != null)
      listener.NewSquaresLeft(squares_left);
  }

  private void IncSquaresLeft() {
    ++squares_left;
    if (listener != null)
      listener.NewSquaresLeft(squares_left);
  }

  private void DecSquaresLeft() {
    --squares_left;
    if (listener != null)
      listener.NewSquaresLeft(squares_left);
  }

  public int GetSquaresLeft() {
    return squares_left;
  }


  // Handle the events
  public void mouseClicked(MouseEvent e) {}

  public void mouseEntered(MouseEvent e) {}

  public void mouseExited(MouseEvent e) {}

  public void mousePressed(MouseEvent e) {

    if ((listener != null) && !user_lock)
      listener.MousePressed();

    // If the user is locked out (ie. end game), ignore update
    if (user_lock || ((mb = MineButton.CurrentButton()) == null) || !mb.IsNumber())
      return;

    // Unhighlight if the wrong buttons are pressed
    if (GetHighlighted())
      UnsetHighlightAt(mb);
    else if ((CountAdj(mb) != mb.GetNumber()) &&
             ((!mb.mouseLeft() && mb.mouseCenter() && !mb.mouseRight()) ||
              (mb.mouseLeft() && !mb.mouseCenter() && mb.mouseRight())))
      SetHighlightAt(mb);

  }

  public void mouseReleased(MouseEvent e) {

    if ((listener != null) && !user_lock)
      listener.MouseReleased();

    if ((user_lock) || (mb == null))
      return;

    if (GetHighlighted())
      UnsetHighlightAt(mb);

    if (mb != MineButton.CurrentButton())
      return;

    boolean hitmine = false;

    // If the user clicked on a free square
    if (mb.IsFree()) {
      //System.out.println(mb.mouseLeft() + " " + mb.mouseCenter() + " " + mb.mouseRight());
      if (mb.mouseLeft() && !mb.mouseCenter() && !mb.mouseRight())
        hitmine = CheckAt(mb);
      else if (!mb.mouseLeft() && !mb.mouseCenter() && mb.mouseRight() && (mine_count > 0))
        FlagAt(mb);
    }

    // Else if the user clicked on a flag
    else if (mb.IsFlag()) {
      if (!mb.mouseLeft() && !mb.mouseCenter() && mb.mouseRight())
        ClearAt(mb);
    }
    else if (((!mb.mouseLeft() && mb.mouseCenter() && !mb.mouseRight()) ||
              (mb.mouseLeft() && !mb.mouseCenter() && mb.mouseRight())) &&
             (CountAdj(mb) == mb.GetNumber()))
      hitmine = ClickAdj(mb);

    if (hitmine) {
      UserDead();
      UserLock();
      ShowAll();
    }

    if (GetSquaresLeft() == 0) {
      UserLock();
      EndGame();
    }
  }

  // This method performs the actual mine checking and accounting
  // It returns true if a mine was hit, otherwise false
  private boolean CheckAt(MineButton target) {

    // The user clicked on a mine
    if (target.IsHiddenMine())
      return true;

    // The user clicked on a safe square
    int count = MineCount(target);
    target.SetNumber(count);
    DecSquaresLeft();

    // If the count came out to zero, check the surrounding FREE squares
    if (count == 0)
      return ClickAdj(target);
    return false;
  }


  // This methods counts the number of mines adjascent to the given target
  private int MineCount(MineButton target) {
    Point 
      min = GetMinBoundary(target.GetID()), 
      max = GetMaxBoundary(target.GetID());
    int count = 0;

    // Count the number of surrounding mines
    for (int y = min.y; y <= max.y; ++y)
      for (int x = min.x; x <= max.x; ++x)
	if (field[x][y].IsHiddenMine())
	  ++count;

    return count;
  }


  // This method places a flag at the given target
  public void FlagAt(MineButton target) {
    target.SetFlag(flag);
    DecSquaresLeft();
    DecMineCount();
  }


  // This method clears a flag from the given target
  public void ClearAt(MineButton target) {
    target.SetFree();
    IncSquaresLeft();
    IncMineCount();
  }


  // This method highlights the adjacent
  // free squares at the given target
  private void SetHighlightAt(MineButton target) {
    Point 
      min = GetMinBoundary(target.GetID()), 
      max = GetMaxBoundary(target.GetID());

    SetHighlighted();

    // Highlight adjacent FREE mines
    for (int y = min.y; y <= max.y; ++y)
      for (int x = min.x; x <= max.x; ++x)
	if (field[x][y].IsFree()) {
	  //System.out.println("(" + x + "," + y + ")");
	  field[x][y].highlight();
	}
  }


  // This method unhighlits the adjacent FREE squares around
  // the given target
  private void UnsetHighlightAt(MineButton target) {
    Point 
      min = GetMinBoundary(target.GetID()), 
      max = GetMaxBoundary(target.GetID());

    UnsetHighlighted();

    // Unhighlight adjacent FREE squares
    for (int y = min.y; y <= max.y; ++y)
      for (int x = min.x; x <= max.x; ++x)
	if (field[x][y].IsFree())
	  field[x][y].unhighlight();
  }


  // These methods provide an interface to the lighted variable
  private void SetHighlighted() {
    lighted = true;
  }

  private void UnsetHighlighted() {
    lighted = false;
  }

  private boolean GetHighlighted() {
    return lighted;
  }


  // This method performs the actual adjacent CheckAts for other methods
  private boolean ClickAdj(MineButton target) {
    Point 
      min = GetMinBoundary(target.GetID()), 
      max = GetMaxBoundary(target.GetID());

    for (int y = min.y; y <= max.y; ++y)
      for (int x = min.x; x <= max.x; ++x)
	if ((field[x][y].IsFree()) && (CheckAt(field[x][y])))
	  return true;

    return false;
  }


  // This method counts and returns the number of adjacent flags
  public int CountAdj(MineButton target) {
    int count = 0;
    Point 
      min = GetMinBoundary(target.GetID()), 
      max = GetMaxBoundary(target.GetID());

    for (int y = min.y; y <= max.y; ++y)
      for (int x = min.x; x <= max.x; ++x)
	if (field[x][y].IsFlag())
	    ++count;

    return count;
  }


  // When the user clicks on a mine, this method shows all of the mines
  // and places WRONG marks where the user placed a flag and there 
  // was no mine
  public void ShowAll() {
    for (int y = 0; y < field_dim.height; ++y)
      for (int x = 0; x < field_dim.width; ++x) {
	if (field[x][y].IsHiddenMine()) {
	  field[x][y].SetMine(mine);
	  field[x][y].Done();
	} else if (field[x][y].IsFlag()) {
	  field[x][y].SetWrong(mineX);
	  field[x][y].Done();
	}
      }
  }


  // This method makes the user know that the game is over
  // All buttons with flags are "flattened"
  private void EndGame() {
    for (int y = 0; y < field_dim.height; ++y)
      for (int x = 0; x < field_dim.width; ++x)
	if (field[x][y].IsFlag())
	    field[x][y].Done();
  }


  // These methods determine the upper-left and lower-right corners of 
  // the rectangle that contains the adjacent squares to the target
  private Point GetMinBoundary(Point target) {
    Point min = new Point(target.x - 1, target.y - 1);

    if (min.x < 0) min.x = 0;
    if (min.y < 0) min.y = 0;

    return min;
  }

  private Point GetMaxBoundary(Point target) {
    Point max = new Point(target.x + 1, target.y + 1);

    if (max.x >= field_dim.width)  max.x = field_dim.width  - 1;
    if (max.y >= field_dim.height) max.y = field_dim.height - 1;

    return max;
  }
}
